@@ -1,24 +1,24 @@
 /* engine-gpgsm.c - GpgSM engine.
-   Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007, 2009,
-                 2010 g10 Code GmbH
-
-   This file is part of GPGME.
-
-   GPGME is free software; you can redistribute it and/or modify it
-   under the terms of the GNU Lesser General Public License as
-   published by the Free Software Foundation; either version 2.1 of
-   the License, or (at your option) any later version.
-
-   GPGME is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.  */
+ * Copyright (C) 2000 Werner Koch (dd9jn)
+ * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007, 2009,
+ *               2010 g10 Code GmbH
+ *
+ * This file is part of GPGME.
+ *
+ * GPGME is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * GPGME is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -82,6 +82,7 @@ struct engine_gpgsm
   iocb_data_t output_cb;
 
   iocb_data_t message_cb;
+  iocb_data_t diag_cb;
 
   struct
   {
@@ -109,6 +110,9 @@ struct engine_gpgsm
   char request_origin[10];
 
   struct gpgme_io_cbs io_cbs;
+
+  /* Memory data containing diagnostics (--logger-fd) of gpgsm */
+  gpgme_data_t diagnostics;
 };
 
 typedef struct engine_gpgsm *engine_gpgsm_t;
@@ -146,6 +150,13 @@ close_notify_handler (int fd, void *opaque)
 	(*gpgsm->io_cbs.remove) (gpgsm->status_cb.tag);
       gpgsm->status_cb.fd = -1;
       gpgsm->status_cb.tag = NULL;
+      /* Because the server keeps on running as long as the
+       * gpgme_ctx_t is valid the diag fd will not receive a close and
+       * thus the operation gets stuck trying to read the diag fd.
+       * The status fd however is closed right after it received the
+       * "OK" from the command.  So we use this event to also close
+       * the diag fd.  */
+      _gpgme_io_close (gpgsm->diag_cb.fd);
     }
   else if (gpgsm->input_cb.fd == fd)
     {
@@ -178,6 +189,13 @@ close_notify_handler (int fd, void *opaque)
       gpgsm->message_cb.fd = -1;
       gpgsm->message_cb.tag = NULL;
     }
+  else if (gpgsm->diag_cb.fd == fd)
+    {
+      if (gpgsm->diag_cb.tag)
+	(*gpgsm->io_cbs.remove) (gpgsm->diag_cb.tag);
+      gpgsm->diag_cb.fd = -1;
+      gpgsm->diag_cb.tag = NULL;
+    }
 }
 
 
@@ -213,6 +231,8 @@ gpgsm_cancel (void *engine)
     _gpgme_io_close (gpgsm->output_cb.fd);
   if (gpgsm->message_cb.fd != -1)
     _gpgme_io_close (gpgsm->message_cb.fd);
+  if (gpgsm->diag_cb.fd != -1)
+    _gpgme_io_close (gpgsm->diag_cb.fd);
 
   if (gpgsm->assuan_ctx)
     {
@@ -234,6 +254,8 @@ gpgsm_release (void *engine)
 
   gpgsm_cancel (engine);
 
+  gpgme_data_release (gpgsm->diagnostics);
+
   free (gpgsm->colon.attic.line);
   free (gpgsm);
 }
@@ -246,17 +268,18 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
   gpgme_error_t err = 0;
   engine_gpgsm_t gpgsm;
   const char *pgmname;
-  const char *argv[5];
+  const char *argv[7];
+  char *diag_fd_str = NULL;
   int argc;
-#if !USE_DESCRIPTOR_PASSING
   int fds[2];
-  int child_fds[4];
-#endif
+  int child_fds[5];
+  int nchild_fds;
   char *dft_display = NULL;
   char dft_ttyname[64];
   char *env_tty = NULL;
   char *dft_ttytype = NULL;
   char *optstr;
+  unsigned int connect_flags;
 
   (void)version; /* Not yet used.  */
 
@@ -284,6 +307,11 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
   gpgsm->message_cb.tag = 0;
   gpgsm->message_cb.server_fd = -1;
   *gpgsm->message_cb.server_fd_str = 0;
+  gpgsm->diag_cb.fd = -1;
+  gpgsm->diag_cb.dir = 1;
+  gpgsm->diag_cb.tag = 0;
+  gpgsm->diag_cb.server_fd = -1;
+  *gpgsm->diag_cb.server_fd_str = 0;
 
   gpgsm->status.fnc = 0;
   gpgsm->colon.fnc = 0;
@@ -300,7 +328,20 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
   gpgsm->io_cbs.event = NULL;
   gpgsm->io_cbs.event_priv = NULL;
 
-#if !USE_DESCRIPTOR_PASSING
+  if (_gpgme_io_pipe (fds, 1) < 0)
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+  gpgsm->diag_cb.fd = fds[0];
+  gpgsm->diag_cb.server_fd = fds[1];
+
+#if USE_DESCRIPTOR_PASSING
+  child_fds[0] = gpgsm->diag_cb.server_fd;
+  child_fds[1] = -1;
+  nchild_fds = 2;
+  connect_flags = ASSUAN_PIPE_CONNECT_FDPASSING;
+#else /*!USE_DESCRIPTOR_PASSING*/
   if (_gpgme_io_pipe (fds, 0) < 0)
     {
       err = gpg_error_from_syserror ();
@@ -328,8 +369,11 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
   child_fds[0] = gpgsm->input_cb.server_fd;
   child_fds[1] = gpgsm->output_cb.server_fd;
   child_fds[2] = gpgsm->message_cb.server_fd;
-  child_fds[3] = -1;
-#endif
+  child_fds[3] = gpgsm->diag_cb.server_fd;
+  child_fds[4] = -1;
+  nchild_fds = 5;
+  connect_flags = 0;
+#endif  /*!USE_DESCRIPTOR_PASSING*/
 
   pgmname = file_name ? file_name : _gpgme_get_default_gpgsm_name ();
 
@@ -340,6 +384,18 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
       argv[argc++] = "--homedir";
       argv[argc++] = home_dir;
     }
+  /* Set up diagnostics */
+  err = gpgme_data_new (&gpgsm->diagnostics);
+  if (err)
+    goto leave;
+  gpgsm->diag_cb.data = gpgsm->diagnostics;
+  argv[argc++] = "--logger-fd";
+  if (gpgrt_asprintf (&diag_fd_str, "%i", gpgsm->diag_cb.server_fd) == -1)
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+  argv[argc++] = diag_fd_str;
   argv[argc++] = "--server";
   argv[argc++] = NULL;
 
@@ -350,26 +406,24 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
     goto leave;
   assuan_ctx_set_system_hooks (gpgsm->assuan_ctx, &_gpgme_assuan_system_hooks);
 
-#if USE_DESCRIPTOR_PASSING
-  err = assuan_pipe_connect (gpgsm->assuan_ctx, pgmname, argv,
-                             NULL, NULL, NULL, ASSUAN_PIPE_CONNECT_FDPASSING);
-#else
   {
-    assuan_fd_t achild_fds[4];
+    assuan_fd_t achild_fds[5];
     int i;
 
     /* For now... */
-    for (i = 0; i < 4; i++)
+    for (i = 0; i < nchild_fds; i++)
       achild_fds[i] = (assuan_fd_t) child_fds[i];
 
     err = assuan_pipe_connect (gpgsm->assuan_ctx, pgmname, argv,
-                               achild_fds, NULL, NULL, 0);
+                               achild_fds, NULL, NULL, connect_flags);
 
-    /* For now... */
-    for (i = 0; i < 4; i++)
+    /* FIXME: Check whether our Windows code still updates the list.*/
+    for (i = 0; i < nchild_fds; i++)
       child_fds[i] = (int) achild_fds[i];
   }
 
+
+#if !USE_DESCRIPTOR_PASSING
   /* On Windows, handles are inserted in the spawned process with
      DuplicateHandle, and child_fds contains the server-local names
      for the inserted handles when assuan_pipe_connect returns.  */
@@ -387,6 +441,8 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
 		sizeof gpgsm->output_cb.server_fd_str, "%d", child_fds[1]);
       snprintf (gpgsm->message_cb.server_fd_str,
 		sizeof gpgsm->message_cb.server_fd_str, "%d", child_fds[2]);
+      snprintf (gpgsm->diag_cb.server_fd_str,
+		sizeof gpgsm->diag_cb.server_fd_str, "%d", child_fds[3]);
     }
 #endif
   if (err)
@@ -499,6 +555,12 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
       goto leave;
     }
 #endif
+  if (!err && _gpgme_io_set_close_notify (gpgsm->diag_cb.fd,
+                                          close_notify_handler, gpgsm))
+    {
+      err = gpg_error (GPG_ERR_GENERAL);
+      goto leave;
+    }
 
  leave:
   /* Close the server ends of the pipes (because of this, we must use
@@ -512,12 +574,14 @@ gpgsm_new (void **engine, const char *file_name, const char *home_dir,
   if (gpgsm->message_cb.server_fd != -1)
     _gpgme_io_close (gpgsm->message_cb.server_fd);
 #endif
+  if (gpgsm->diag_cb.server_fd != -1)
+    _gpgme_io_close (gpgsm->diag_cb.server_fd);
 
   if (err)
     gpgsm_release (gpgsm);
   else
     *engine = gpgsm;
-
+  free (diag_fd_str);
   return err;
 }
 
@@ -836,7 +900,7 @@ status_handler (void *opaque, int fd)
 	{
 	  /* Try our best to terminate the connection friendly.  */
 	  /*	  assuan_write_line (gpgsm->assuan_ctx, "BYE"); */
-          TRACE3 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: error from assuan (%d) getting status line : %s",
                   fd, err, gpg_strerror (err));
 	}
@@ -848,7 +912,7 @@ status_handler (void *opaque, int fd)
 	    err = atoi (&line[4]);
 	  if (! err)
 	    err = gpg_error (GPG_ERR_GENERAL);
-          TRACE2 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: ERR line - mapped to: %s",
                   fd, err ? gpg_strerror (err) : "ok");
 	  /* Try our best to terminate the connection friendly.  */
@@ -877,7 +941,7 @@ status_handler (void *opaque, int fd)
               gpgsm->colon.any = 0;
               err = gpgsm->colon.fnc (gpgsm->colon.fnc_value, NULL);
             }
-          TRACE2 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: OK line - final status: %s",
                   fd, err ? gpg_strerror (err) : "ok");
 	  _gpgme_io_close (gpgsm->status_cb.fd);
@@ -952,7 +1016,7 @@ status_handler (void *opaque, int fd)
 		    dst++;
 		}
 	    }
-          TRACE2 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: D line; final status: %s",
                   fd, err? gpg_strerror (err):"ok");
         }
@@ -994,7 +1058,7 @@ status_handler (void *opaque, int fd)
               linelen -= nwritten;
             }
 
-          TRACE2 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: D inlinedata; final status: %s",
                   fd, err? gpg_strerror (err):"ok");
         }
@@ -1032,7 +1096,7 @@ status_handler (void *opaque, int fd)
 	    }
 	  else
 	    fprintf (stderr, "[UNKNOWN STATUS]%s %s", line + 2, rest);
-          TRACE3 (DEBUG_CTX, "gpgme:status_handler", gpgsm,
+          TRACE (DEBUG_CTX, "gpgme:status_handler", gpgsm,
 		  "fd 0x%x: S line (%s) - final status: %s",
                   fd, line+2, err? gpg_strerror (err):"ok");
 	}
@@ -1062,8 +1126,8 @@ add_io_cb (engine_gpgsm_t gpgsm, iocb_data_t *iocbd, gpgme_io_cb_t handler)
 {
   gpgme_error_t err;
 
-  TRACE_BEG2 (DEBUG_ENGINE, "engine-gpgsm:add_io_cb", gpgsm,
-              "fd %d, dir %d", iocbd->fd, iocbd->dir);
+  TRACE_BEG  (DEBUG_ENGINE, "engine-gpgsm:add_io_cb", gpgsm,
+              "fd=%d, dir %d", iocbd->fd, iocbd->dir);
   err = (*gpgsm->io_cbs.add) (gpgsm->io_cbs.add_priv,
 			      iocbd->fd, iocbd->dir,
 			      handler, iocbd->data, &iocbd->tag);
@@ -1139,6 +1203,8 @@ start (engine_gpgsm_t gpgsm, const char *command)
     err = add_io_cb (gpgsm, &gpgsm->output_cb, _gpgme_data_inbound_handler);
   if (!err && gpgsm->message_cb.fd != -1)
     err = add_io_cb (gpgsm, &gpgsm->message_cb, _gpgme_data_outbound_handler);
+  if (!err && gpgsm->diag_cb.fd != -1)
+    err = add_io_cb (gpgsm, &gpgsm->diag_cb, _gpgme_data_inbound_handler);
 
   if (!err)
     err = assuan_write_line (gpgsm->assuan_ctx, command);
@@ -1341,11 +1407,12 @@ set_recipients_from_string (engine_gpgsm_t gpgsm, const char *string)
 {
   gpg_error_t err = 0;
   char *line = NULL;
-  int no_pubkey = 0;
+  int ignore = 0;
+  int any = 0;
   const char *s;
   int n;
 
-  for (;;)
+  do
     {
       while (*string == ' ' || *string == '\t')
         string++;
@@ -1360,25 +1427,32 @@ set_recipients_from_string (engine_gpgsm_t gpgsm, const char *string)
       while (n && (string[n-1] == ' ' || string[n-1] == '\t'))
         n--;
 
-      gpgrt_free (line);
-      if (gpgrt_asprintf (&line, "RECIPIENT %.*s", n, string) < 0)
+      if (!ignore && n == 2 && !memcmp (string, "--", 2))
+        ignore = 1;
+      else if (!ignore && n > 2 && !memcmp (string, "--", 2))
+        err = gpg_error (GPG_ERR_UNKNOWN_OPTION);
+      else if (n) /* Not empty - use it.  */
         {
-          err = gpg_error_from_syserror ();
-          break;
+          gpgrt_free (line);
+          if (gpgrt_asprintf (&line, "RECIPIENT %.*s", n, string) < 0)
+            err = gpg_error_from_syserror ();
+          else
+            {
+              err = gpgsm_assuan_simple_command (gpgsm, line, gpgsm->status.fnc,
+                                                 gpgsm->status.fnc_value);
+              if (!err)
+                any = 1;
+            }
         }
+
       string += n + !!s;
-
-      err = gpgsm_assuan_simple_command (gpgsm, line, gpgsm->status.fnc,
-					 gpgsm->status.fnc_value);
-
-      /* Fixme: Improve error reporting.  */
-      if (gpg_err_code (err) == GPG_ERR_NO_PUBKEY)
-	no_pubkey++;
-      else if (err)
-        break;
     }
+  while (!err);
+
+  if (!err && !any)
+    err = gpg_error (GPG_ERR_MISSING_KEY);
   gpgrt_free (line);
-  return err? err : no_pubkey? gpg_error (GPG_ERR_NO_PUBKEY) : 0;
+  return err;
 }
 
 
@@ -1392,7 +1466,7 @@ gpgsm_encrypt (void *engine, gpgme_key_t recp[], const char *recpstring,
 
   if (!gpgsm)
     return gpg_error (GPG_ERR_INV_VALUE);
-  if (!recp)
+  if (!recp && !recpstring) /* Symmetric only */
     return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
 
   if ((flags & GPGME_ENCRYPT_NO_ENCRYPT_TO))
@@ -1438,6 +1512,12 @@ gpgsm_export (void *engine, const char *pattern, gpgme_export_mode_t mode,
   if (!gpgsm)
     return gpg_error (GPG_ERR_INV_VALUE);
 
+  if ((mode & ~(GPGME_EXPORT_MODE_SECRET
+                |GPGME_EXPORT_MODE_MINIMAL
+                |GPGME_EXPORT_MODE_RAW
+                |GPGME_EXPORT_MODE_PKCS12)))
+    return gpg_error (GPG_ERR_NOT_SUPPORTED);
+
   if (!pattern)
     pattern = "";
 
@@ -1485,6 +1565,12 @@ gpgsm_export_ext (void *engine, const char *pattern[], gpgme_export_mode_t mode,
   if (!gpgsm)
     return gpg_error (GPG_ERR_INV_VALUE);
 
+  if ((mode & ~(GPGME_EXPORT_MODE_SECRET
+                |GPGME_EXPORT_MODE_MINIMAL
+                |GPGME_EXPORT_MODE_RAW
+                |GPGME_EXPORT_MODE_PKCS12)))
+    return gpg_error (GPG_ERR_NOT_SUPPORTED);
+
   if (pattern && *pattern)
     {
       const char **pat = pattern;
@@ -1622,16 +1708,24 @@ gpgsm_genkey (void *engine,
 
 
 static gpgme_error_t
-gpgsm_import (void *engine, gpgme_data_t keydata, gpgme_key_t *keyarray)
+gpgsm_import (void *engine, gpgme_data_t keydata, gpgme_key_t *keyarray,
+              const char *keyids[], const char *import_filter,
+              const char *key_origin)
 {
   engine_gpgsm_t gpgsm = engine;
   gpgme_error_t err;
   gpgme_data_encoding_t dataenc;
   int idx;
 
+  (void)import_filter;
+  (void)key_origin;
+
   if (!gpgsm)
     return gpg_error (GPG_ERR_INV_VALUE);
 
+  if (keyids)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
   if (keydata && keyarray)
     return gpg_error (GPG_ERR_INV_VALUE); /* Only one is allowed.  */
 
@@ -1661,7 +1755,7 @@ gpgsm_import (void *engine, gpgme_data_t keydata, gpgme_key_t *keyarray)
               && *keyarray[idx]->subkeys->fpr)
             buflen += strlen (keyarray[idx]->subkeys->fpr) + 1;
         }
-      /* Allocate a bufer with extra space for the trailing Nul
+      /* Allocate a buffer with extra space for the trailing Nul
          introduced by the use of stpcpy.  */
       buffer = malloc (buflen+1);
       if (!buffer)
@@ -2030,11 +2124,19 @@ gpgsm_verify (void *engine, gpgme_data_t sig, gpgme_data_t signed_text,
   err = gpgsm_set_fd (gpgsm, INPUT_FD, map_data_enc (gpgsm->input_cb.data));
   if (err)
     return err;
-  if (plaintext)
+  if (!signed_text)
     {
       /* Normal or cleartext signature.  */
-      gpgsm->output_cb.data = plaintext;
-      err = gpgsm_set_fd (gpgsm, OUTPUT_FD, 0);
+      if (plaintext)
+        {
+          gpgsm->output_cb.data = plaintext;
+          err = gpgsm_set_fd (gpgsm, OUTPUT_FD, 0);
+        }
+      else
+        {
+          /* No output requested.  */
+          gpgsm_clear_fd (gpgsm, OUTPUT_FD);
+        }
       gpgsm_clear_fd (gpgsm, MESSAGE_FD);
     }
   else
@@ -2061,12 +2163,40 @@ gpgsm_getauditlog (void *engine, gpgme_data_t output, unsigned int flags)
   engine_gpgsm_t gpgsm = engine;
   gpgme_error_t err = 0;
 
+
   if (!gpgsm || !output)
     return gpg_error (GPG_ERR_INV_VALUE);
 
-  if ((flags & GPGME_AUDITLOG_DIAG))
+  if ((flags & GPGME_AUDITLOG_DIAG) && (flags & GPGME_AUDITLOG_HTML))
     return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
 
+  if ((flags & GPGME_AUDITLOG_DIAG))
+    {
+      char buf[BUFFER_SIZE];
+      int nread;
+      int any_written = 0;
+      gpgme_data_rewind (gpgsm->diagnostics);
+
+      while ((nread = gpgme_data_read (gpgsm->diagnostics,
+                                       buf, BUFFER_SIZE)) > 0)
+        {
+          any_written = 1;
+          if (gpgme_data_write (output, buf, nread) == -1)
+            return gpg_error_from_syserror ();
+        }
+      if (!any_written)
+        return gpg_error (GPG_ERR_NO_DATA);
+
+      if (nread == -1)
+        return gpg_error_from_syserror ();
+
+      gpgme_data_rewind (output);
+      return 0;
+    }
+
+  if (!gpgsm->assuan_ctx)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
 #if USE_DESCRIPTOR_PASSING
   gpgsm->output_cb.data = output;
   err = gpgsm_set_fd (gpgsm, OUTPUT_FD, 0);
@@ -2140,7 +2270,7 @@ gpgsm_io_event (void *engine, gpgme_event_io_t type, void *type_data)
 {
   engine_gpgsm_t gpgsm = engine;
 
-  TRACE3 (DEBUG_ENGINE, "gpgme:gpgsm_io_event", gpgsm,
+  TRACE (DEBUG_ENGINE, "gpgme:gpgsm_io_event", gpgsm,
           "event %p, type %d, type_data %p",
           gpgsm->io_cbs.event, type, type_data);
   if (gpgsm->io_cbs.event)
@@ -2212,11 +2342,12 @@ struct engine_ops _gpgme_engine_ops_gpgsm =
     gpgsm_keylist_ext,
     NULL,               /* keylist_data */
     NULL,               /* keysign */
+    NULL,               /* revsig */
     NULL,               /* tofu_policy */
     gpgsm_sign,
-    NULL,		/* trustlist */
     gpgsm_verify,
     gpgsm_getauditlog,
+    NULL,               /* setexpire */
     NULL,               /* opassuan_transact */
     NULL,		/* conf_load */
     NULL,		/* conf_save */
