@@ -984,8 +984,8 @@ make_db_file_name (const char *issuer_hash)
 
 
 /* Hash the file FNAME and return the MD5 digest in MD5BUFFER. The
-   caller must allocate MD%buffer wityh at least 16 bytes. Returns 0
-   on success. */
+ * caller must allocate MD5buffer with at least 16 bytes. Returns 0
+ * on success. */
 static int
 hash_dbfile (const char *fname, unsigned char *md5buffer)
 {
@@ -1660,6 +1660,7 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   size_t n;
   gcry_sexp_t s_sig = NULL, s_hash = NULL, s_pkey = NULL;
   unsigned int saltlen = 0;  /* (used only with use_pss)  */
+  int pkalgo;
 
   /* This also stops debugging on the MD.  */
   gcry_md_final (md);
@@ -1787,6 +1788,54 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
                              gcry_md_read (md, algo),
                              saltlen);
     }
+  else if ((pkalgo = pk_algo_from_sexp (s_pkey)) == GCRY_PK_ECC)
+    {
+      unsigned int qbits0, qbits;
+
+      qbits0 = gcry_pk_get_nbits (s_pkey);
+      qbits = qbits0 == 521? 512 : qbits0;
+
+      if ((qbits%8))
+	{
+	  log_error ("ECDSA requires the hash length to be a"
+                     " multiple of 8 bits\n");
+	  err = gpg_error (GPG_ERR_INTERNAL);
+          goto leave;
+	}
+
+      /* Don't allow any Q smaller than 160 bits.  */
+      if (qbits < 160)
+	{
+	  log_error (_("%s key uses an unsafe (%u bit) hash\n"),
+                     gcry_pk_algo_name (pkalgo), qbits0);
+	  err = gpg_error (GPG_ERR_INTERNAL);
+          goto leave;
+	}
+
+      /* Check if we're too short.  */
+      n = gcry_md_get_algo_dlen (algo);
+      if (n < qbits/8)
+        {
+	  log_error (_("a %u bit hash is not valid for a %u bit %s key\n"),
+                     (unsigned int)n*8,
+                     qbits0,
+                     gcry_pk_algo_name (pkalgo));
+          if (n < 20)
+            {
+              err = gpg_error (GPG_ERR_INTERNAL);
+              goto leave;
+            }
+        }
+
+      /* Truncate.  */
+      if (n > qbits/8)
+        n = qbits/8;
+
+      err = gcry_sexp_build (&s_hash, NULL, "(data(flags raw)(value %b))",
+                             (int)n,
+                             gcry_md_read (md, algo));
+
+    }
   else
     {
       err = gcry_sexp_build (&s_hash, NULL,
@@ -1804,7 +1853,7 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   /* Pass this on to the signature verification. */
   err = gcry_pk_verify (s_sig, s_hash, s_pkey);
   if (DBG_X509)
-    log_debug ("gcry_pk_verify: %s\n", gpg_strerror (err));
+    log_debug ("%s: gcry_pk_verify: %s\n", __func__, gpg_strerror (err));
 
  leave:
   xfree (sigval);
@@ -2614,21 +2663,21 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
   ksba_name_t distpoint = NULL;
   ksba_name_t issuername = NULL;
   char *distpoint_uri = NULL;
-  char *issuername_uri = NULL;
   int any_dist_point = 0;
   int seq;
+  gpg_error_t last_err = 0;
 
   /* Loop over all distribution points, get the CRLs and put them into
      the cache. */
   if (opt.verbose)
     log_info ("checking distribution points\n");
   seq = 0;
-  while ( !(err = ksba_cert_get_crl_dist_point (cert, seq++,
+  while (xfree (distpoint), xfree (issuername),
+         !(err = ksba_cert_get_crl_dist_point (cert, seq++,
                                                 &distpoint,
                                                 &issuername, NULL )))
     {
       int name_seq;
-      gpg_error_t last_err = 0;
 
       if (!distpoint && !issuername)
         {
@@ -2639,13 +2688,11 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
                     suitable CRL. */
         }
 
-      xfree (issuername_uri); issuername_uri = NULL;
-
       /* Get the URIs.  We do this in a loop to iterate over all names
          in the crlDP. */
       for (name_seq=0; ksba_name_enum (distpoint, name_seq); name_seq++)
         {
-          xfree (distpoint_uri); distpoint_uri = NULL;
+          xfree (distpoint_uri);
           distpoint_uri = ksba_name_get_uri (distpoint, name_seq);
           if (!distpoint_uri)
             continue;
@@ -2669,6 +2716,7 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
 
           if (opt.verbose)
             log_info ("fetching CRL from '%s'\n", distpoint_uri);
+          crl_close_reader (reader);
           err = crl_fetch (ctrl, distpoint_uri, &reader);
           if (err)
             {
@@ -2688,28 +2736,16 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
               last_err = err;
               continue; /* with the next name. */
             }
-          last_err = 0;
-          break; /* Ready. */
+          goto leave; /* Ready - we got the CRL. */
         }
-      if (last_err)
-        {
-          err = last_err;
-          goto leave;
-        }
-
-      ksba_name_release (distpoint); distpoint = NULL;
-
-      /* We don't do anything with issuername_uri yet but we keep the
-         code for documentation. */
-      issuername_uri =  ksba_name_get_uri (issuername, 0);
-      ksba_name_release (issuername); issuername = NULL;
-
-      /* Close the reader.  */
-      crl_close_reader (reader);
-      reader = NULL;
     }
   if (gpg_err_code (err) == GPG_ERR_EOF)
     err = 0;
+  if (!err && last_err)
+    {
+      err = last_err;
+      goto leave;
+    }
 
   /* If we did not found any distpoint, try something reasonable. */
   if (!any_dist_point )
@@ -2717,9 +2753,6 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
       if (opt.verbose)
         log_info ("no distribution point - trying issuer name\n");
 
-      crl_close_reader (reader);
-      reader = NULL;
-
       issuer = ksba_cert_get_issuer (cert, 0);
       if (!issuer)
         {
@@ -2730,6 +2763,7 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
 
       if (opt.verbose)
         log_info ("fetching CRL from default location\n");
+      crl_close_reader (reader);
       err = crl_fetch_default (ctrl, issuer, &reader);
       if (err)
           {
@@ -2752,7 +2786,6 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
  leave:
   crl_close_reader (reader);
   xfree (distpoint_uri);
-  xfree (issuername_uri);
   ksba_name_release (distpoint);
   ksba_name_release (issuername);
   ksba_free (issuer);
