@@ -423,23 +423,33 @@ keyserver_any_configured (ctrl_t ctrl)
 }
 
 int
-keyserver_import_keyid (u32 *keyid, void *dummy, int quick)
+keyserver_import_keyid (u32 *keyid, void *dummy, unsigned int flags)
 {
   (void)keyid;
   (void)dummy;
-  (void)quick;
+  (void)flags;
   return -1;
 }
 
 int
 keyserver_import_fprint (ctrl_t ctrl, const byte *fprint,size_t fprint_len,
-			 struct keyserver_spec *keyserver, int quick)
+			 struct keyserver_spec *keyserver, unsigned int flags)
 {
   (void)ctrl;
   (void)fprint;
   (void)fprint_len;
   (void)keyserver;
-  (void)quick;
+  (void)flags;
+  return -1;
+}
+
+int
+keyserver_import_fprint_ntds (ctrl_t ctrl,
+                              const byte *fprint, size_t fprint_len)
+{
+  (void)ctrl;
+  (void)fprint;
+  (void)fprint_len;
   return -1;
 }
 
@@ -451,19 +461,19 @@ keyserver_import_cert (const char *name)
 }
 
 gpg_error_t
-keyserver_import_wkd (ctrl_t ctrl, const char *name, int quick,
+keyserver_import_wkd (ctrl_t ctrl, const char *name, unsigned int flags,
                       unsigned char **fpr, size_t *fpr_len)
 {
   (void)ctrl;
   (void)name;
-  (void)quick;
+  (void)flags;
   (void)fpr;
   (void)fpr_len;
   return GPG_ERR_BUG;
 }
 
 int
-keyserver_import_name (const char *name,struct keyserver_spec *spec)
+keyserver_import_mbox (const char *name,struct keyserver_spec *spec)
 {
   (void)name;
   (void)spec;
@@ -534,12 +544,14 @@ get_override_session_key (DEK *dek, const char *string)
 
 /* Stub: */
 int
-decrypt_data (ctrl_t ctrl, void *procctx, PKT_encrypted *ed, DEK *dek)
+decrypt_data (ctrl_t ctrl, void *procctx, PKT_encrypted *ed, DEK *dek,
+              int *compliance_error)
 {
   (void)ctrl;
   (void)procctx;
   (void)ed;
   (void)dek;
+  (void)compliance_error;
   return GPG_ERR_GENERAL;
 }
 
@@ -569,13 +581,14 @@ check_secret_key (PKT_public_key *pk, int n)
  */
 DEK *
 passphrase_to_dek (int cipher_algo, STRING2KEY *s2k, int create, int nocache,
-                   const char *tmp, int *canceled)
+                   const char *tmp, unsigned int flags, int *canceled)
 {
   (void)cipher_algo;
   (void)s2k;
   (void)create;
   (void)nocache;
   (void)tmp;
+  (void)flags;
 
   if (canceled)
     *canceled = 0;
@@ -799,3 +812,11 @@ get_revocation_reason (PKT_signature *sig, char **r_reason,
     *r_comment = NULL;
   return 0;
 }
+
+const char *
+impex_filter_getval (void *cookie, const char *propname)
+{
+  (void)cookie;
+  (void)propname;
+  return NULL;
+}
