@@ -71,7 +71,7 @@ wks_set_status_fd (int fd)
 }
 
 
-/* Write a status line with code NO followed by the outout of the
+/* Write a status line with code NO followed by the output of the
  * printf style FORMAT.  The caller needs to make sure that LFs and
  * CRs are not printed.  */
 void
@@ -121,7 +121,7 @@ append_to_uidinfo_list (uidinfo_list_t *list, const char *uid, time_t created,
   strcpy (sl->uid, plainuid);
   sl->created = created;
   sl->flags = 0;
-  sl->mbox = mailbox_from_userid (plainuid);
+  sl->mbox = mailbox_from_userid (plainuid, 0);
   sl->expired = !!expired;
   sl->revoked = !!revoked;
   sl->next = NULL;
@@ -246,7 +246,7 @@ wks_get_key (estream_t *r_key, const char *fingerprint, const char *addrspec,
   ccparray_put (&ccp, "--always-trust");
   if (!binary)
     ccparray_put (&ccp, "--armor");
-  ccparray_put (&ccp, "--export-options=export-clean");
+  ccparray_put (&ccp, "--export-options=export-minimal");
   ccparray_put (&ccp, "--export-filter");
   ccparray_put (&ccp, filterexp);
   ccparray_put (&ccp, "--export");
@@ -953,7 +953,7 @@ wks_fname_from_userid (const char *userid, int hash_only,
   if (r_addrspec)
     *r_addrspec = NULL;
 
-  addrspec = mailbox_from_userid (userid);
+  addrspec = mailbox_from_userid (userid, 0);
   if (!addrspec)
     {
       if (opt.verbose || hash_only)
@@ -1155,7 +1155,7 @@ install_key_from_spec_file (const char *fname)
   char *line = NULL;
   size_t linelen = 0;
   size_t maxlen = 2048;
-  char *fields[2];
+  const char *fields[2];
   unsigned int lnr = 0;
 
   if (!fname || !strcmp (fname, ""))
@@ -1263,7 +1263,7 @@ wks_cmd_install_key (const char *fname, const char *userid)
   if (!userid)
     return install_key_from_spec_file (fname);
 
-  addrspec = mailbox_from_userid (userid);
+  addrspec = mailbox_from_userid (userid, 0);
   if (!addrspec)
     {
       log_error ("\"%s\" is not a proper mail address\n", userid);
@@ -1272,8 +1272,7 @@ wks_cmd_install_key (const char *fname, const char *userid)
     }
 
   if (!classify_user_id (fname, &desc, 1)
-      && (desc.mode == KEYDB_SEARCH_MODE_FPR
-          || desc.mode == KEYDB_SEARCH_MODE_FPR20))
+      && desc.mode == KEYDB_SEARCH_MODE_FPR)
     {
       /* FNAME looks like a fingerprint.  Get the key from the
        * standard keyring.  */
